@@ -18,10 +18,6 @@
 ;; You should have received a copy of the GNU General Public License
 ;; along with this program.  If not, see <https://www.gnu.org/licenses/>.
 
-;; URL: https://github.com/yyoncho/dap-mode
-;; Package-Requires: ((emacs "25.1") (dash "2.14.1") (lsp-mode "4.0"))
-;; Version: 0.2
-
 ;;; Commentary:
 ;; Adapter for https://github.com/erlang-ls/erlang_ls
 ;; Created by Roberto Aloi (@robertoaloi)
@@ -33,10 +29,9 @@
 (defun dap-erlang--populate-start-file-args (conf)
   "Populate CONF with the required arguments."
   (-> conf
-    (plist-put :dap-server-path '("els_dap"))
-    (plist-put :projectDir (lsp-find-session-folder (lsp-session) (buffer-file-name)))
-    (plist-put :cwd (lsp-find-session-folder (lsp-session) (buffer-file-name)))
-    ))
+      (plist-put :dap-server-path '("els_dap"))
+      (plist-put :projectDir (lsp-find-session-folder (lsp-session) (buffer-file-name)))
+      (plist-put :cwd (lsp-find-session-folder (lsp-session) (buffer-file-name)))))
 (dap-register-debug-provider "erlang" 'dap-erlang--populate-start-file-args)
 
 (provide 'dap-erlang)
